/*-------------------------------------------------------------------------+
| Copyright 2012 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package test.org.fortiss.af3.rcp.application.semantic.alignment.java;

import static java.io.File.separator;
import static org.conqat.lib.commons.filesystem.FileSystemUtils.deleteRecursively;
import static org.conqat.lib.commons.filesystem.FileSystemUtils.ensureDirectoryExists;
import static org.fortiss.af3.generator.common.utils.JavaLanguageModelElementFactory.SRC_GEN_SOURCE_PACKAGE;
import static org.fortiss.tooling.base.utils.SystemUtils.getTempDirPath;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.tools.JavaCompiler;
import javax.tools.ToolProvider;

import org.fortiss.af3.component.generator.java.ComponentArchitectureToJavaSourcePackageTransformation;
import org.fortiss.af3.component.generator.java.ComponentProgramToJavaSourcePackageTransformation;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.expression.generator.java.DataDictionaryToJavaSourcePackageTransformation;
import org.fortiss.af3.expression.language.evaluation.NoVal;
import org.fortiss.af3.expression.model.DataDictionary;
import org.fortiss.af3.expression.utils.ExpressionModelElementFactory;
import org.fortiss.af3.generator.common.model.java.JavaSourcePackage;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.af3.project.model.typesystem.ITerm;
import org.fortiss.af3.project.utils.ProjectUtils;
import org.fortiss.af3.rcp.application.AF3ApplicationActivator;
import org.fortiss.af3.testing.model.TestCase;
import org.fortiss.af3.testing.model.TestInput;
import org.fortiss.af3.testing.model.TestOutput;
import org.fortiss.af3.testing.model.TestStep;
import org.fortiss.af3.testing.model.TestSuite;
import org.fortiss.af3.testing.model.TestingModelFactory;
import org.fortiss.af3.testing.simulator.TestCaseSimulator;
import org.fortiss.af3.testing.simulator.TestCaseSimulatorException;
import org.fortiss.af3.testing.ui.Af3TestingUIActivator;
import org.fortiss.af3.testing.utils.TestingModelElementFactory;
import org.fortiss.tooling.kernel.extension.exception.TransformationFailedException;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.utils.IdentifierUtils;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

/**
 * Java code generator test. Checks using the Simple Traffic Lights model that the
 * generated Java code behaves equivalently to the model simulator.
 * 
 * @author trachtenherz
 */
public class JavaEquivalenceTest_SimpleTrafficLights {

	/** Plug-in id */
	public static final String PLUGIN_ID = AF3ApplicationActivator.PLUGIN_ID;

	/** Project */
	public FileProject project;

	/** Component architecture of the model. */
	ComponentArchitecture compArch;
	/** Root component of the model. */
	Component rootComp;
	/** Data dictionary. */
	DataDictionary dataDict;

	/** Class representing the root component in the generated Java code */
	protected Class<?> gen_rootComp;

	/** Class representing the data dictionary in the generated Java code */
	protected Class<?> gen_dataDict;

	/** Method for clearing root component's input ports in the generated Java code */
	Method genMethod_clear_inputs;
	/** Method for clearing root component's output ports in the generated Java code */
	Method genMethod_clear_outputs;
	/** Method for initializing the root component in the generated Java code */
	Method genMethod_init;
	/** Method for performing a step of the root component in the generated Java code */
	Method genMethod_perform_step;

	/** Input port tlarchInButtonA in the generated Java code. */
	Field genField_InButtonA;
	/** Input port tlarchOutTrafficSignal in the generated Java code. */
	Field genField_OutTrafficSignal;
	/** Input port tlarchOutPedestrianSignal in the generated Java code. */
	Field genField_OutPedestrianSignal;
	/** Input port tlarchOutIndicatorSignalA in the generated Java code. */
	Field genField_OutIndicatorSignalA;
	/** Input port tlarchOutIndicatorSignalB in the generated Java code. */
	Field genField_OutIndicatorSignalB;

	/** Map of constant names to methods representing the constants in the generated Java code */
	Map<String, Method> gen_dataDictMap = new HashMap<String, Method>();

	/** Test data. */
	enum InputSignal {
		/** Test data value */
		Present;
	}

	/** String value of the {@code Present} signal for the InButtonA input port. */
	public static final String INPUT_SIGNAL_VALUE_STRING = "Present";

	/** File representing the target forlder for Java code generation. */
	File targetLocation;
	/** Path to the root folder of the generated Java code. */
	public File javaSourcePath;
	/** Name of Java class representing the root component of the model. */
	public String rootCompClassName;

	/**
	 * Set up.
	 * 
	 * @throws MalformedURLException
	 *             if the temporary user directory for code generation cannot be found.
	 */
	@Before
	public void setup() throws MalformedURLException {
		setupProject();
		setupJavaPaths();
		generateJavaCode();
		setupGeneratedJavaCode(javaSourcePath, rootCompClassName);
	}

	/** Loads the Traffic Lights model project for test. */
	public void setupProject() {
		project = ProjectUtils.loadProjectFromPlugin(Af3TestingUIActivator.PLUGIN_ID,
				"/test-data/SimpleTrafficLightsExample.af3_23", true);

		for(IProjectRootElement element : project.getRootElements()) {
			if(element instanceof ComponentArchitecture) {
				compArch = (ComponentArchitecture)element;
				rootComp = compArch.getTopComponent();
			} else if(element instanceof DataDictionary) {
				dataDict = (DataDictionary)element;
			}
		}
	}

	/**
	 * Sets up the paths to generated Java code.
	 * 
	 * @throws MalformedURLException
	 *             if the temporary user directory for code generation cannot be found.
	 * 
	 * @see #javaSourcePath
	 * @see #rootCompClassName
	 */
	protected void setupJavaPaths() throws MalformedURLException {
		rootCompClassName = IdentifierUtils.getUniqueIdentifier(rootComp);

		targetLocation = new File(getTempDirPath(false, getClass().getName()));
		javaSourcePath = new File(targetLocation, SRC_GEN_SOURCE_PACKAGE);
	}

	/**
	 * Generates the Java code for the loaded model {@link #setupProject()} to a temporary folder.
	 */
	protected void generateJavaCode() {
		try {
			JavaSourcePackage javaSourcePackage =
					(JavaSourcePackage)(new ComponentArchitectureToJavaSourcePackageTransformation())
							.transform(compArch, null);
			javaSourcePackage.setBaseLocation("");

			if(targetLocation.exists()) {
				deleteRecursively(targetLocation);
			}
			ensureDirectoryExists(targetLocation);
			javaSourcePackage.writeToLocation(targetLocation.getAbsolutePath());
		} catch(TransformationFailedException e) {
			e.printStackTrace();
			throw new AssertionError(e);
		} catch(IOException e) {
			e.printStackTrace();
			throw new AssertionError(e);
		}
	}

	/** Sets up the generated Java code. */
	protected void setupGeneratedJavaCode(File sourcePath, String rootCompClassName) {
		// Compile the generated Java code
		String mainJavaFilePath = sourcePath + separator + "af3" + separator + "javagen" +
				separator + "component" + separator + rootCompClassName + ".java";
		String[] args = new String[] {"-sourcepath", sourcePath.toString(), mainJavaFilePath};
		JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
		compiler.run(null, null, null, args);

		ClassLoader classLoader = null;
		try {
			// Due to a bug in the URLClassPath (see
			// https://github.com/AdoptOpenJDK/openjdk-jdk11/blob/999dbd4192d0f819cb5224f26e9e7fa75ca6f289/src/java.base/share/classes/jdk/internal/loader/URLClassPath.java#L471)
			// generated classes are not found on Windows. Hence, we must "unify" the source path.
			URL sourcePathURL = new URL(sourcePath.toURI().toString().replace("\\", "/"));
			classLoader = URLClassLoader.newInstance(new URL[] {sourcePathURL});
			// Load the classes for root component and data dictionary
			gen_rootComp = Class.forName(
					ComponentProgramToJavaSourcePackageTransformation.COMPONENT_JAVA_PACKAGE + "." +
							rootCompClassName,
					true, classLoader);
			gen_dataDict = Class.forName(
					DataDictionaryToJavaSourcePackageTransformation.DD_JAVA_PACKAGE + "." +
							DataDictionaryToJavaSourcePackageTransformation.DD_JAVA_CLASS,
					true, classLoader);
		} catch(Exception e) {
			e.printStackTrace();
			throw new AssertionError(e);
		}

		// Store references to certain methods and fields in the generated Java code
		readGenMethods();
		readGenFields();
		readGenDataDict();
	}

	/** Stored method references from the generated Java code required to perform the test. */
	protected void readGenMethods() {
		Method[] methods = gen_rootComp.getDeclaredMethods();
		for(Method m : methods) {
			final String methodName = m.getName();
			if(methodName.contains("clear_inputs")) {
				genMethod_clear_inputs = m;
			} else if(methodName.contains("clear_outputs")) {
				genMethod_clear_outputs = m;
			} else if(methodName.contains("init")) {
				genMethod_init = m;
			} else if(methodName.contains("perform_step")) {
				genMethod_perform_step = m;
			} else {
				// Nothing
			}
		}
	}

	/** Stored field references from the generated Java code required to perform the test. */
	protected void readGenFields() {
		Field[] fields = gen_rootComp.getDeclaredFields();
		for(Field f : fields) {
			final String fieldName = f.getName();
			if(fieldName.contains("InButtonA")) {
				genField_InButtonA = f;
			} else if(fieldName.contains("OutTrafficSignal")) {
				genField_OutTrafficSignal = f;
			} else if(fieldName.contains("OutPedestrianSignal")) {
				genField_OutPedestrianSignal = f;
			} else if(fieldName.contains("OutIndicatorSignalA")) {
				genField_OutIndicatorSignalA = f;
			} else if(fieldName.contains("OutIndicatorSignalB")) {
				genField_OutIndicatorSignalB = f;
			} else {
				// Nothing
			}
		}
	}

	/** Stores the data dictionary from the generated Java code. */
	protected void readGenDataDict() {
		Method[] methods = gen_dataDict.getDeclaredMethods();
		for(Method m : methods) {
			gen_dataDictMap.put(m.getName(), m);
		}
	}

	/** Performs one step for the generated Java code. */
	protected void genPerformStep()
			throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		genMethod_clear_outputs.invoke(null);
		genMethod_perform_step.invoke(null);
		genMethod_clear_inputs.invoke(null);
	}

	/**
	 * @param inputSignals
	 *            Input values for the tlarchInButtonA input port of the Traffic Lights Example.
	 *            {@link InputSignal#Present} correspond to {@code Present} message inputs while
	 *            {@code null} values correspond to {@code NoVal} inputs.
	 * @param testSuite
	 *            {@link TestSuite} to which the test case will be added. If {@code null} a new
	 *            {@link TestSuite} instance is created.
	 * @return
	 * 		A {@link TestCase} with inputs corresponding to {@code inputSignals}.
	 */
	protected TestCase generateTestCase(InputSignal[] inputSignals, TestSuite testSuite) {
		TestCase testCase = TestingModelFactory.eINSTANCE.createTestCase();
		if(testSuite == null) {
			testSuite = TestingModelElementFactory.createTestSuite("Java Generator Test Suite",
					rootComp);
		}
		testSuite.getTestCases().add(testCase);

		List<TestStep> stepList = testCase.getTestSteps();

		for(InputSignal in : inputSignals) {
			TestStep testStep = TestingModelFactory.eINSTANCE.createTestStep();
			TestInput testInput = TestingModelFactory.eINSTANCE.createTestInput();
			ITerm value = in == InputSignal.Present
					? ExpressionModelElementFactory.funcCall("Present") : NoVal.NOVAL.toITerm();
			testInput.setValue(value);
			testStep.getInputs().add(testInput);

			stepList.add(testStep);
		}
		return testCase;
	}

	/**
	 * Returns the current output port values for the generated java code.
	 * 
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	protected Object[] genOutputResult() throws IllegalArgumentException, IllegalAccessException {
		return new Object[] {genField_OutTrafficSignal.get(null),
				genField_OutPedestrianSignal.get(null), genField_OutIndicatorSignalA.get(null),
				genField_OutIndicatorSignalB.get(null)};
	}

	/**
	 * Simulates the model and the generated Java code with the specified input data. After each
	 * simulation step the equivalence of the results from the model and generated code is asserted
	 * ({@link Assert#assertTrue(String, boolean)})
	 * 
	 * @see #simOutputsEquiv(TestStep, Object[])
	 * 
	 * @param inputSignals
	 *            Input values. The i-th value is {@link InputSignal#Present} if the input in the
	 *            i-th step should be the {@code Present} signal and {@code null} if the input
	 *            should be {@code NoVal}.
	 */
	protected void assertSimulationEquiv(InputSignal[] inputSignals)
			throws IllegalArgumentException, IllegalAccessException {
		TestCase testCase = generateTestCase(inputSignals, null);
		TestCaseSimulator sim = new TestCaseSimulator(testCase, rootComp);
		try {
			sim.simulate(true);
		} catch(TestCaseSimulatorException e) {
			e.printStackTrace();
			throw new AssertionError(e);
		}
		List<TestStep> testSteps = testCase.getTestSteps();

		try {
			genMethod_init.invoke(null);
			for(int i = 0; i < inputSignals.length; i++) {
				if(inputSignals[i] == InputSignal.Present) {
					genField_InButtonA.set(null,
							gen_dataDictMap.get(INPUT_SIGNAL_VALUE_STRING).invoke(null));
				}
				// Perform one step in the generated Java code
				genPerformStep();
				// Compare the results from the model simulation and the generated Java code
				// execution
				assertTrue("Simulation step " + i,
						simOutputsEquiv(testSteps.get(i), genOutputResult()));
			}
		} catch(InvocationTargetException e) {
			e.getCause().printStackTrace();
		}
	}

	/**
	 * Checks that the simulation results of the model (stored in {@code testStep}) and of the
	 * generated Java code stored in {@code genCodeOutput} are equivalent.
	 */
	protected boolean simOutputsEquiv(TestStep testStep, Object[] genCodeOutput) {
		List<TestOutput> testOutputs = testStep.getOutputs();
		for(int i = 0; i < testOutputs.size(); i++) {
			ITerm simOutValue = testOutputs.get(i).getSimulatedValue();
			Object genCodeValue = genCodeOutput[i];

			String simOutStr = simOutValue.toString();
			// Remove parentheses from string representation
			int parPos = simOutStr.indexOf('(');
			if(parPos >= 0) {
				simOutStr = simOutStr.substring(0, parPos).trim();
			}
			// Result string for generated Java code
			String genCodeStr = genCodeValue == null ? NoVal.NOVAL_STRING : genCodeValue.toString();
			/*
			 * We consider outputs equivalent if the result string from Java code execution starts
			 * with the string representing the simulated model output (possible parentheses and
			 * spaces have been removed before comparison). The result string in the Java code may
			 * contain additional information like a unique id so that we check not for equality but
			 * for prefix relation.
			 */
			if(!genCodeStr.startsWith(simOutStr)) {
				// Outputs are not equivalent
				return false;
			}
		}
		return true;
	}

	/**
	 * Tests behavior equivalence for an input sequence containing one {@code Present} signal,
	 * 
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	@Test
	public void test1() throws IllegalArgumentException, IllegalAccessException {
		final int testDataSize = 20;
		InputSignal[] testData1 = new InputSignal[testDataSize];
		testData1[2] = InputSignal.Present;
		assertSimulationEquiv(testData1);
	}

	/**
	 * Tests behavior equivalence for an input sequence containing several {@code Present} signals
	 * with distance less than time needed to completely process one {@code Present} input signal.
	 * 
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	@Test
	public void test2() throws IllegalArgumentException, IllegalAccessException {
		final int testDataSize = 40;
		InputSignal[] testData1 = new InputSignal[testDataSize];
		testData1[10] = testData1[20] = testData1[30] = InputSignal.Present;
		assertSimulationEquiv(testData1);
	}

	/**
	 * Tests behavior equivalence for a long input sequence containing periodic {@code Present}
	 * signals.
	 * 
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	@Test
	public void test3() throws IllegalArgumentException, IllegalAccessException {
		final int testDataSize = 1000;
		final int inputPeriod = 10;
		InputSignal[] testData1 = new InputSignal[testDataSize];
		for(int i = 0; i < testDataSize / inputPeriod; ++i) {
			testData1[i * inputPeriod] = InputSignal.Present;
		}
		assertSimulationEquiv(testData1);
	}
}
